#include "typedef.h"

#include <assert.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <malloc.h>
#include <fcntl.h>
#include <stdio.h>
#include <string.h>

#include "stringcache.h"
#include "util.h"
#include "object.h"
#include "opcode.h"
#include "class.h"
#include "sort.h"

#if defined(WIN32)
#include <io.h>
#include <direct.h>

#define open _open
#define read _read
#define close _close
#define alloca _alloca
#define getcwd _getcwd
#else
#include <wchar.h>
#include <errno.h>
#include <unistd.h>
#include <alloca.h>

#define _O_BINARY 0
#endif

extern int				g_dynamic;
extern int				g_verbose;
extern char				*g_classpathP;
extern char				*g_java[10];
extern char				**g_envPP;

extern FILE				*stdoutF;

static u1T				*g_imageP         = 0;
static const u1T		*g_image_atP      = 0;
static int				g_image_size      = 0;
static int				g_image_max       = -1;

static linenumberT		*g_linenumbersP    = 0;
static int				g_linenumbers_max = 0;

#define Align(X) ((((int) (X)) + 3) & ~3)

#define AssertIndex(X) assert(X >= 0 && X < m_constant_pool_count)
#define AssertCode(X)  assert(X >= 0 && X < code_length)
#define AssertEndCode(X)  assert(X >= 0 && X <= code_length)



static const u1T	*
getU2(const u1T *startP, u2T *valP)
{
	register	const u1T	*P = startP;
	register	u2T			val;

	assert(!(((int) valP) & 1));
	val   = (*P++ << 8);
	val  |= *P++;
	*valP = val;
	return(P);
}

static const u1T	*
getU4(const u1T *startP, u4T *valP)
{
	register	const u1T	*P = startP;
	register	u4T			val;

	assert(!(((int) valP) & 3));
	val   = (*P++ << 24);
	val  |= (*P++ << 16);
	val  |= (*P++ << 8);
	val  |= *P++;
	*valP = val;
	return (P);
}
	
static const u1T *
getU8(const u1T *startP, u8T *valP)
{
	register	const u1T	*P = startP;

	u8T	val;

//	assert(!(((int) valP) & 7));
	val   = (((u8T) *P) << 56); ++P;
	val  |= (((u8T) *P) << 48); ++P;
	val  |= (((u8T) *P) << 40); ++P;
	val  |= (((u8T) *P) << 32); ++P;
	val  |= (((u8T) *P) << 24); ++P;
	val  |= (((u8T) *P) << 16); ++P;
	val  |= (((u8T) *P) << 8);  ++P;
	val  |= *P++;
	*valP = val;
	return (P);
}

/*
  4.4.7 The CONSTANT_Utf8_info Structure

  The CONSTANT_Utf8_info structure is used to represent constant string values. UTF-8 strings are encoded so that character 
  sequences that contain only non-null ASCII characters can be represented using only 1 byte per character, but characters 
  of up to 16 bits can be represented. All characters in the range '\u0001' to '\u007F' are represented by a single byte: 

0  bits 6-0  

The 7 bits of data in the byte give the value of the character represented. 
The null character ('\u0000') and characters in the range '\u0080' to '\u07FF' are represented by a pair of bytes x and y:

x:  1  1  0  bits 10-6  
y:  1  0  bits 5-0  

The bytes represent the character with the value ((x & 0x1f) << 6) + (y & 0x3f). 

Characters in the range '\u0800' to '\uFFFF' are represented by 3 bytes x, y, and z:

x:  1  1  1  0  bits 15-12  
y:  1  0  bits 11-6  
z:  1  0  bits 5-0  

The character with the value ((x & 0xf) << 12) + ((y & 0x3f) << 6) + (z & 0x3f) is represented by the bytes. 

The bytes of multibyte characters are stored in the class file in big-endian (high byte first) order.

*/

static const u1T *
getUtf8(int bytes, const u1T *compressedP, u2T **stringPP, int load)
{
	const u1T	*P       = compressedP;
	const u1T	*endP    = P + bytes;
	u2T			*startP  = *stringPP;
	u2T			*stringP = startP;
	u1T			c;
	u2T			val;

	for (; P < endP; ++P) {
		c = *P;
		if (!(c & 0x80)) {
			// One byte encoding
			val = c;
		} else {
			if (!(c & 0x20)) {
				// Two byte encoding
				val  = (c & 0x1F) << 6;
			} else {
				// Three byte encoding
				val  = (c & 0xF) << 12;
				c    = *++P;
				val |= (c & 0x3F) << 6;
			}
			c    = *++P;
			val |= (c & 0x3F);
		}
		if (load) {
			*stringP = val;
		}
		++stringP;
	}
	if (load) {
		*stringP = 0;
	}
	++stringP;
	*stringPP  = stringP;
	assert(P == endP);
	return(endP);
}

/* Read the class image from disk */

static int
read_image(const char *filenameP, struct stat *statsP)
{
	int				fileno;
	int				ret;

	g_image_size = statsP->st_size;
	if (g_image_size > g_image_max) {
		if (g_imageP) {
			free(g_imageP);
		}
		g_imageP     = (u1T *) xmalloc(g_image_size + 1);
		g_image_max  = g_image_size;
	}
	fileno = open(filenameP, O_RDONLY | _O_BINARY, 0);
	if (fileno == -1) {
		fprintf(stderr, "Can't open %s\n", filenameP);
		goto fail;
	}
	ret = read(fileno, g_imageP, g_image_size);
	if (ret != g_image_size) {
		fprintf(stderr, "Can't read %s\n", filenameP);
		goto fail;
	}

	if (close(fileno)) {
		fprintf(stderr, "Can't close %s\n", filenameP);
		goto fail;
	}
	g_imageP[g_image_size] = 0;
	return(0);
fail:
	return(-1);
}

/*
    ClassFile {
    	u4 magic;
    	u2 minor_version;
    	u2 major_version;
    	u2 constant_pool_count;
    	cp_info constant_pool[constant_pool_count-1];
    	u2 access_flags;
    	u2 this_class;
    	u2 super_class;
    	u2 interfaces_count;
    	u2 interfaces[interfaces_count];
    	u2 fields_count;
    	field_info fields[fields_count];
    	u2 methods_count;
    	method_info methods[methods_count];
    	u2 attributes_count;
    	attribute_info attributes[attributes_count];
    }
*/

static int			g_memory         = 0;
static u1T			*g_memory_startP = 0;
static int			g_extra          = 0;
static u1T			*g_extra_startP  = 0;
static int			g_string_need    = 0;
static const u1T	*g_image_endP    = 0;

int
Cclass::load_constant_pool(const char *filenameP, int load)
{
	u1T				*memoryP;
	u1T				*stringP, *startP;
	cp_infoT		*cpP, *endP;
	constantE		tag;
	int				i, need;
	u4T				magic;
	u2T				length;

	g_image_atP = g_imageP;
	if (!load) {
		memoryP     = (u1T *) 0;
		stringP     = (u1T *) 0;
		g_image_atP = getU4(g_image_atP, &magic);

		if (magic != 0xCAFEBABE) {
			fprintf(stderr, "%s is not a valid class file - has magic value 0x%08x\n", filenameP, magic);
			goto fail;
		}
		g_image_atP   = getU2(g_image_atP, &m_minor_version);
		g_image_atP   = getU2(g_image_atP, &m_major_version);
		g_image_atP   = getU2(g_image_atP, &m_constant_pool_count);
	} else {
		memoryP      = (u1T *) m_constant_poolP;
		stringP      = (u1T *) (m_constant_poolP + m_constant_pool_count);
		g_image_atP += 10;
	}

	cpP = (cp_infoT *) memoryP;
	if (m_constant_pool_count) {

		if (load) {
			cpP->tag = CONSTANT_Null;
		}
		++cpP;
		for (i = 1; i < m_constant_pool_count; ++i) {
			tag = (constantE) *g_image_atP++;
			if (load) {
				cpP->tag = tag;
			}
			switch (tag) {
			case CONSTANT_Utf8:
				g_image_atP  = getU2(g_image_atP, &length);
				startP       = stringP;
				g_image_atP  = getUtf8(length, g_image_atP, (u2T **) &stringP, load);
				if (load) {
					cpP->u.stringP = stringCache((u2T *) startP);
				}
				break;
			case CONSTANT_Integer:
			case CONSTANT_Float:
				if (load) {
					g_image_atP     = getU4(g_image_atP, &cpP->u.u4val);
				} else {
					g_image_atP    += 4;
				}
				break;
			case CONSTANT_NameAndType:
				if (load) {
					g_image_atP     = getU2(g_image_atP, &cpP->u.ref.index1);
					AssertIndex(cpP->u.ref.index1);
					g_image_atP     = getU2(g_image_atP, &cpP->u.ref.index2);
					AssertIndex(cpP->u.ref.index2);
				} else {
					g_image_atP    += 4;
				}
				break;
			case CONSTANT_Long:
			case CONSTANT_Double:
				if (load) {
					g_image_atP     = getU8(g_image_atP, &cpP->u.u8val);
				} else {
					g_image_atP    += 8;
				}
				// 8 byte values take 2 slots in the constant pool table
				++cpP;
				if (load) {
					cpP->tag = CONSTANT_Null;
				}
				++i;
				break;
			case CONSTANT_Class:
			case CONSTANT_String:
				if (load) {
					g_image_atP     = getU2(g_image_atP, &cpP->u.ref.index1);
					AssertIndex(cpP->u.ref.index1);
				} else {
					g_image_atP    += 2;
				}
				break;
			case CONSTANT_Fieldref:
			case CONSTANT_Methodref:
			case CONSTANT_InterfaceMethodref:
				if (load) {
					g_image_atP     = getU2(g_image_atP, &cpP->u.ref.index1);
					AssertIndex(cpP->u.ref.index1);
					g_image_atP     = getU2(g_image_atP, &cpP->u.ref.index2);
					AssertIndex(cpP->u.ref.index2);
				} else {
					g_image_atP    += 4;
				}
				break;
			default:
				fprintf(stderr, "Unknown constant code %d\n", tag);
				goto fail;
			}
			++cpP;
		}
		assert(cpP == m_constant_poolP + m_constant_pool_count);
		memoryP = (u1T *) cpP;
	}
	if (!load) {
		g_image_endP     = g_image_atP;
		g_memory         = memoryP - (u1T *) 0;
		g_string_need    = stringP - (u1T *) 0;
		need             = g_memory + g_string_need;
		memoryP          = (u1T *) xmalloc(need + strlen(filenameP) + 1);
		m_constant_poolP = (cp_infoT *) memoryP;
		m_filenameP      = (char *) (memoryP + need);
		strcpy(m_filenameP, filenameP);
	} else {
		assert((memoryP - ((u1T *) m_constant_poolP)) == g_memory);
		assert((stringP - ((u1T *) m_constant_poolP)) == g_memory + g_string_need);
		assert(g_image_atP == g_image_endP);

		cpP  = m_constant_poolP;
		endP = cpP + m_constant_pool_count;
		while (++cpP < endP) {
			switch (cpP->tag) {
			case CONSTANT_String:
			case CONSTANT_Class:
				// Simplify the indirection here up front to help in managing constants later
				// Do it after the entire table is loaded to anticipate forward references if
				// they should exist
				cpP->u.stringP = get_Utf8(cpP->u.ref.index1);
				break;
		}	}
	}
	return(0);
fail:
	return(-1);
}

#define HASH_SIZE 0x190758

static pcodeT	**hash_table[HASH_SIZE] = {0};

static void
hash_pcode(pcodeT *pcodeP)
{
	int			 pcode;
	unsigned int hash;
	unsigned int hash1;
	pcodeT		 **hashPP;

	pcodeP->m_hashP = 0;
	pcode           = pcodeP->m_pcode;
	hash            = opcodes[pcode].hash;

	assert(hash != 0xFFFFFF);

	switch (pcode) {
	case ILOAD:		// 0x000000
	case LLOAD:		// 0x010000
	case FLOAD:		// 0x020000
	case DLOAD:		// 0x030000
	case ALOAD:		// 0x040000
	case ISTORE:	// 0x050000
	case LSTORE:	// 0x060000
	case FSTORE:	// 0x070000
	case DSTORE:	// 0x080000
	case ASTORE:	// 0x090000
	{
		varT	*varP = (varT *) pcodeP->m_tableP;
		
		if (varP->m_nameP) {
			hash1 = ((unsigned int) varP->m_nameP) >> 2;
		} else {
			hash1 = varP->m_number;
		}
		hash |= (hash1 & 0xFFFF);
		break;
	}
	case GETSTATIC:	// 0x0a0000	
	case GETFIELD:	// 0x0b0000
	case PUTSTATIC: // 0x0c0000
	case PUTFIELD:	// 0x0d0000
	{
		fieldrefT	*fieldrefP = (fieldrefT *) pcodeP->m_tableP;

		hash1 = (((unsigned int) fieldrefP->m_classnameP) ^ 
			     ((unsigned int) fieldrefP->m_nameP)
			    ) >> 2;
		hash |= (hash1 & 0xFFFF);
		break;
	}
	case INVOKEVIRTUAL:	// 0x0e0000
	case INVOKESPECIAL:	// 0x0f0000
	case INVOKESTATIC:	// 0x100000
	{
		methodrefT	*methodrefP = (methodrefT *) pcodeP->m_tableP;

		hash1     = (((unsigned int) methodrefP->m_classnameP) ^ 
			         ((unsigned int) methodrefP->m_nameP)      ^
					 ((unsigned int) methodrefP->m_descriptorP)
					) >> 2;
		hash |= (hash1 & 0xFFFF);
		break;
	}
	case INVOKEINTERFACE: // 0x110000
	{
		interfacerefT	*methodrefP = (interfacerefT *) pcodeP->m_tableP;

		hash1     = (((unsigned int) methodrefP->m_classnameP) ^ 
			         ((unsigned int) methodrefP->m_nameP)      ^
					 ((unsigned int) methodrefP->m_descriptorP)
					 ) >> 2;
		hash1 ^= methodrefP->m_args;
		hash  |= (hash1 & 0xFFFF);
		break;
	}
	case NEW:		// 0x120000
	case ANEWARRAY:	// 0x130000
	case CHECKCAST:	// 0x140000
	case INSTANCEOF:// 0x150000
	{
		classrefT	*classrefP;
		
		classrefP = (classrefT *) pcodeP->m_tableP;

		hash1  = ((unsigned int) classrefP->m_classnameP) >> 2;
		hash  |= (hash1 & 0xFFFF);
		break;
	}
	case MULTIANEWARRAY: // 0x160000
	{
		arrayrefT	*arrayrefP = (arrayrefT *) pcodeP->m_tableP;

		hash1  = ((unsigned int) arrayrefP->m_classnameP) >> 2;
		hash1 ^= arrayrefP->m_dimensions;
		hash  |= (hash1 & 0xFFFF);
		break;
	}
	case IINC:		// 0x17000
	{
		iincT	*varP = (iincT *) pcodeP->m_tableP;
		
		if (varP->m_nameP) {
			hash1     = ((unsigned int) varP->m_nameP) >> 2;
		} else {
			hash1     = varP->m_number;
		}
		hash |= ((hash1 + varP->m_s4) & 0xFFFF);
		break;
	}
	case LDC:
	case LDC_W:
	case LDC2_W:	// 0x180000
	{
		cp_infoT	*cpP = (cp_infoT *) pcodeP->m_tableP;

		switch (cpP->tag) {
		case CONSTANT_Integer:
			hash1 = cpP->u.ival & 0x3FFF;
			break;
		case CONSTANT_Float:
			hash1 = 0x4000 | (cpP->u.ival & 0x3FFF);
			break;
		case CONSTANT_Double:
		case CONSTANT_Long:
			hash1 = 0x8000 | ((cpP->u.ival ^ (unsigned int) cpP->u.lval) & 0x3FFF);
			break;
		case CONSTANT_String:
		case CONSTANT_Class:
		{
			const u2T *P;

			hash1 = 0;
			for (P = cpP->u.stringP; *P; ++P) {
				hash1 <<= 1;
				hash1  ^= *P;
			}
			hash1 = 0xc000 | (hash1 & 0x3FFF);
			break;
		}
		default:
			assert(0);
		}
		hash |= hash1;
		break;
	}
	case IFEQ:		// 0x190000 - 0x19003F
	case IFNE:		
	case IFLT:		
	case IFGE:		
	case IFGT:
	case IFLE:
	case IF_ICMPEQ:
	case IF_ICMPNE:
	case IF_ICMPLT:
	case IF_ICMPGE:
	case IF_ICMPGT:
	case IF_ICMPLE:
	case IF_ACMPEQ:
	case IF_ACMPNE:
	case IFNULL:
	case IFNOTNULL:
	case GOTO:
	case GOTO_W:
	case JSR:
	case JSR_W:
	case RET_W:
	case LABEL:
		if (pcodeP->m_s4 < 0) {
			hash |= 0x20;
		}
		break;
	case BIPUSH:	// 0x190400
	case SIPUSH:	// 0x190500
	case NEWARRAY:	// 0x190600
		hash |= pcodeP->m_u4 & 0xFF;
		break;
	}

	assert(hash >= 0 && hash < HASH_SIZE);

	if (hashPP = hash_table[hash]) {
		*hashPP = pcodeP; 
	}
	hash_table[hash] = &(pcodeP->m_hashP);
}

const static u2T	sourceFile[] =
	{'S','o','u','r','c','e','F','i','l','e',0};
const static u2T	constantValue[] =
	{'C','o','n','s','t','a','n','t','V','a','l','u','e',0};
const static u2T	code[] =
	{'C','o','d','e',0};
const static u2T	deprecated[] =
	{'D','e','p','r','e','c','a','t','e','d',0};
const static u2T	enclosingMethod[] = 
	{'E','n','c','l','o','s','i','n','g','M','e','t','h','o','d',0};
const static u2T	exceptions[] = 
	{'E','x','c','e','p','t','i','o','n','s',0};
const static u2T	innerClasses[] =
	{'I','n','n','e','r','C','l','a','s','s','e','s',0};
const static u2T	lineNumberTable[] =
	{'L','i','n','e','N','u','m','b','e','r','T','a','b','l','e',0};
const static u2T	localVariableTable[] =
	{'L','o','c','a','l','V','a','r','i','a','b','l','e','T','a','b','l','e',0};
const static u2T	synthetic[] =
	{'S','y','n','t','h','e','t','i','c',0};

int
Cclass::load_tables(int load)
{
	static	unsigned int		g_pcode_count;

	const u1T		*image_atP   = g_image_atP;
	const u1T		*image_at1P;
	const u1T		*image_at2P;
	u1T				*memoryP, *extraP;
	field_infoT		*fieldP, *field_endP;
	method_infoT	*methodP, *method_endP;
	u2T				attribute_name_index, throws_count;
	u2T				*nameP, *interfaceP, *interface_endP;
	u4T				method_number, i, j;
	u4T				attribute_length, index;
	u2T				fields_count, methods_count;
	u2T				attributes_count;
	labelT			*sort_atP, *sort_endP;
	
	/*
		u2 access_flags;
    	u2 this_class;
    	u2 super_class;
    	u2 interfaces_count;
	*/
	if (load) {
		sort_atP      = Csort::g_labelsP;
		sort_endP     = sort_atP + Csort::g_labels_count;
		memoryP       = g_memory_startP;
		extraP        = g_extra_startP;
		m_interfacesP = (u2T *) memoryP;
		image_atP     = getU2(image_atP, &m_access_flags);
		image_atP     = getU2(image_atP, &m_this_class);
		AssertIndex(m_this_class);
		image_atP     = getU2(image_atP, &m_super_class);
		AssertIndex(m_super_class);
		image_atP    += 2;

		nameP = get_classname(m_this_class);
		if (!m_nameP) {
			m_nameP       = nameP;
			index         = hash(nameP);
			m_hashP       = g_hash[index];
			g_hash[index] = this;
		} else {
			if (unicodecmp(nameP, m_nameP)) {
				fprintf(stderr,"Class \"%s\" has internal classname \"%s\"\n", m_nameP, nameP);
		}	}
		*g_tailPP = this;
		g_tailPP  = &m_nextP;
	} else {
		memoryP    = 0;
		extraP     = 0;
		image_atP += 6;
		image_atP  = getU2(image_atP, &m_interfaces_count);
		Csort::clear_sort();
	}

	// interfaces

	interfaceP     = (u2T *) memoryP;
	interface_endP = interfaceP + m_interfaces_count;	
	if (load) {
		for (; interfaceP < interface_endP; ++interfaceP) {
			image_atP = getU2(image_atP, interfaceP);
			AssertIndex(*interfaceP);
		} 
	} else {
		image_atP += 2 * m_interfaces_count;
	}
	memoryP   = (u1T *) interface_endP;

	// fields 

	image_atP = getU2(image_atP, &fields_count);
	m_fields_count = fields_count;
	m_fieldsP  = fieldP = (field_infoT *) Align(memoryP);
	field_endP = fieldP + fields_count;
	memoryP    = (u1T *) field_endP;

	for (; fieldP < field_endP; ++fieldP) {
		if (load) {
			image_atP = getU2(image_atP, &fieldP->m_access_flags);
			image_atP = getU2(image_atP, &fieldP->m_name_index);			// get_Utf8(fieldsP->name_index);
			AssertIndex(fieldP->m_name_index);
			image_atP = getU2(image_atP, &fieldP->m_descriptor_index);	// get_Utf8(fieldsP->descriptor_index);
			AssertIndex(fieldP->m_descriptor_index);
			image_atP = getU2(image_atP, &attributes_count);
			fieldP->m_constant_index = 0;
			fieldP->m_flags          = 0;
		} else {
			image_atP += 6;
			image_atP = getU2(image_atP, &attributes_count);
		}
		for (i = 0; i < attributes_count; ++i) {
			image_atP  = getU2(image_atP, &attribute_name_index);
			AssertIndex(attribute_name_index);
			image_atP  = getU4(image_atP, &attribute_length);
			image_at1P = image_atP;
			image_atP += attribute_length;

			if (load) {
				nameP     = get_Utf8(attribute_name_index);
				if (!unicodecmp(nameP, constantValue)) {
					assert(attribute_length == 2);
					getU2(image_at1P, &fieldP->m_constant_index);
					AssertIndex(fieldP->m_constant_index);
					fieldP->m_flags |= ConstantX;
					continue;
				}
				if (!unicodecmp(nameP, deprecated)) {
					fieldP->m_flags |= DeprecatedX;
					continue;
			}	}
		}
	}

	// methodsP

	image_atP = getU2(image_atP, &methods_count);
	m_methods_count = methods_count;
	m_methodsP      = methodP = (method_infoT *) Align(memoryP);
	method_endP     = methodP + methods_count;
	memoryP         = (u1T *) method_endP;
	for (method_number = 0; methodP < method_endP; ++methodP, ++method_number) {

		if (load) {
			memset(methodP, 0, sizeof(method_infoT));
			methodP->m_classP = this;
			image_atP = getU2(image_atP, &methodP->m_access_flags);
			image_atP = getU2(image_atP, &methodP->m_name_index);
			AssertIndex(methodP->m_name_index);
			image_atP = getU2(image_atP, &methodP->m_descriptor_index);
			AssertIndex(methodP->m_descriptor_index);
			image_atP = getU2(image_atP, &attributes_count);
		} else {
			image_atP += 6;
			image_atP = getU2(image_atP, &attributes_count);
		}
		for (i = 0; i < attributes_count; ++i) {
			image_atP  = getU2(image_atP, &attribute_name_index);
			AssertIndex(attribute_name_index);
			image_atP  = getU4(image_atP, &attribute_length);
			nameP      = get_Utf8(attribute_name_index);

			image_at1P = image_atP;
			image_atP += attribute_length;
			if (load) {
				if (!unicodecmp(nameP, synthetic)) {
					methodP->m_flags |= SyntheticX;
					continue;
				}
				if (!unicodecmp(nameP, deprecated)) {
					methodP->m_flags |= DeprecatedX;
					continue;
			}	}

			if (!unicodecmp(nameP, exceptions)) {
				image_at1P = getU2(image_at1P, &throws_count);
				memoryP    = (u1T *) Align(memoryP);
				if (load) {
					methodP->m_throws_count  = throws_count;
					methodP->m_throws_indexP = (u2T *) memoryP;
				}
				if (load) {
					for (j = 0; j < throws_count; ++j) {
						image_at1P = getU2(image_at1P, (u2T *) memoryP);
						AssertIndex(*((u2T *) memoryP));
						memoryP   += 2;
					}
				} else {
					memoryP += (2 * throws_count);
				}
				continue;
			}
			if (!unicodecmp(nameP, code)) {
				localvariableT	*localvariablesP = 0;
				localvariableT	*localvariable_endP = 0;
				localvariableT	*localvariableP;
				u2T				localvariables_count;

				u2T				u2, exception_table_count;
				s2T				s2;
				u4T				u4, code_length;
				s4T				s4;
				const u1T		*startCodeP, *endCodeP;
				u1T				u1;
				u4T				attributes_count, lineno_changes_at_pc, current_lineno, pc, flags;
				pcodeE			pcode;
				const opcodeT	*opcodeP;
				u4T				wide_index = 0;
				exceptionT		*exceptionP, *exception_endP;
				u4T				attribute_length;
				const u2T		*nameP;
				u2T				attribute_name_index;
				linenumberT		*linenumberP, *linenumber_endP;
				pcodeT			*pcodeP;

				if (load) {
					image_at1P  = getU2(image_at1P, &methodP->m_max_stack);
					image_at1P  = getU2(image_at1P, &methodP->m_max_locals);
				} else {
					image_at1P += 4;
				}

				current_lineno       = 0xFFFFFFFF;
				lineno_changes_at_pc = 0x7FFFFFFF;
				linenumberP          = 0;
				
				startCodeP     = getU4(image_at1P, &code_length);
				endCodeP       = image_at1P = startCodeP + code_length;

				image_at1P     = getU2(image_at1P, &exception_table_count);
				exceptionP     = (exceptionT *) Align(memoryP);
				exception_endP = exceptionP + exception_table_count;

				if (load) {
					methodP->m_exceptions_count = exception_table_count;
					methodP->m_exceptionsP      = exceptionP;

					for (; exceptionP < exception_endP; ++exceptionP) {
						image_at1P = getU2(image_at1P, &exceptionP->m_start_pc);
						AssertCode(exceptionP->m_start_pc);
						image_at1P = getU2(image_at1P, &exceptionP->m_end_pc);
						AssertEndCode(exceptionP->m_end_pc);
						image_at1P = getU2(image_at1P, &exceptionP->m_handler_pc);
						AssertCode(exceptionP->m_handler_pc);
						image_at1P = getU2(image_at1P, &exceptionP->m_catch_type);
						AssertIndex(exceptionP->m_catch_type);
					}
				} else {
					image_at1P += 8 * exception_table_count;
				}
				memoryP = (u1T *) exception_endP;

				image_at1P = getU2(image_at1P, &u2);
				attributes_count = u2;
				linenumberP      = 0;
				current_lineno   = 0;
				for (j = 0; j < attributes_count; ++j) {
					image_at1P = getU2(image_at1P, &attribute_name_index);
					AssertIndex(attribute_name_index);
					nameP  = get_Utf8(attribute_name_index);
					image_at1P  = getU4(image_at1P, &attribute_length);
					image_at2P  = image_at1P;
					image_at1P += attribute_length;
						
					if (load && !unicodecmp(nameP, lineNumberTable)) {
						u2T			linenumbers_count;

						image_at2P      = getU2(image_at2P, &linenumbers_count);
						if (linenumbers_count >= g_linenumbers_max) {
							g_linenumbers_max = (linenumbers_count + 1024) & ~1023;
							if (g_linenumbersP) {
								xfree((void **) &g_linenumbersP);
							}
							g_linenumbersP = (linenumberT *) xmalloc(g_linenumbers_max * sizeof(linenumberT));
						}

						linenumberP     = g_linenumbersP;
						linenumber_endP = g_linenumbersP + linenumbers_count;

						for (; linenumberP < linenumber_endP; ++linenumberP) {
							image_at2P = getU2(image_at2P, &linenumberP->m_start_pc);
							image_at2P = getU2(image_at2P, &linenumberP->m_line_number);
						}
						assert(image_at2P == image_at1P);
														
						linenumber_endP->m_line_number = -1;
						linenumber_endP->m_start_pc    = 0xFFFF;
						current_lineno                 = g_linenumbersP->m_line_number;
						linenumberP                    = g_linenumbersP+1;
						lineno_changes_at_pc           = linenumberP->m_start_pc;
						continue;
					}

					if (!unicodecmp(nameP, localVariableTable)) {

						image_at2P         = getU2(image_at2P, &localvariables_count);
						localvariablesP    = (localvariableT *) Align(memoryP);
						localvariable_endP = localvariablesP + localvariables_count;

						if (load) {
							for (localvariableP = localvariablesP; localvariableP < localvariable_endP; ++localvariableP) {
								image_at2P = getU2(image_at2P, &localvariableP->m_start_pc);
								image_at2P = getU2(image_at2P, &u2);
								localvariableP->m_end_pc = localvariableP->m_start_pc + u2;
								image_at2P = getU2(image_at2P, &u2);
								localvariableP->m_nameP = get_Utf8(u2);
								image_at2P = getU2(image_at2P, &u2);
								localvariableP->m_descriptorP = get_Utf8(u2);
								image_at2P = getU2(image_at2P, &localvariableP->m_index);
								AssertCode(localvariableP->m_start_pc);
								AssertEndCode((unsigned int) (localvariableP->m_end_pc));
							}
							methodP->m_localvariablesP      = localvariablesP;
							methodP->m_localvariables_count = localvariables_count;
							assert (image_at2P == image_at1P);
						} 
						memoryP = (u1T *) localvariable_endP;
						continue;
					}
				}

				memoryP = (u1T *) Align(memoryP);
				memoryP += sizeof(method_infoT *);
				pcodeP  = (pcodeT *) memoryP;
				if (load) {
					((method_infoT **) pcodeP)[-1] = methodP;
					methodP->m_pcodeP = pcodeP;
				}
				
				for (image_at2P = startCodeP; image_at2P < endCodeP; ++pcodeP) {
					pc = (u4T) (image_at2P - startCodeP);
					if (linenumberP) {
						while (pc >= lineno_changes_at_pc) {
							current_lineno = linenumberP->m_line_number;
							++linenumberP;
							lineno_changes_at_pc = linenumberP->m_start_pc;
					}	}

					pcode   = (pcodeE) *image_at2P++;
					opcodeP = opcodes + pcode;
					flags   = opcodeP->flags;

					if (flags & INVALIDX) {
						fprintf(stderr, ": Unknown pcode operation of %d 0x%02x\n", code);
						goto fail;
					}

					s4 = opcodeP->variable;
					u4 = (u4T) s4;

					if (flags & (U1X|U2X|U4X)) {
						if ((flags & U1X)) {
							if (wide_index) {
								image_at2P = getU2(image_at2P, &u2);
								u4         = u2;
								s4         = ((signed short) u2);
								if (pcode != IINC) {
									wide_index = 0;
								}
							} else {
								u1 = *image_at2P++;
								u4 = u1;
								s4 = ((signed char) u1);
							}
						} else if (flags & U2X) {
							image_at2P = getU2(image_at2P, &u2);
							u4         = u2;
							s4         = ((signed short) u2);
						} else {
							image_at2P = getU4(image_at2P, &u4);
							s4         = (signed int) u4;
					}	}

					if (load) {

						for (;sort_atP < sort_endP && sort_atP->method == method_number && sort_atP->to <= pc; ++sort_atP) {
							pcodeP->m_lineno  = current_lineno;
							pcodeP->m_pc      = sort_atP->to;
							pcodeP->m_offset  = (pcodeP - (pcodeT *) memoryP);
							pcodeP->m_pcode   = sort_atP->pcode;
							pcodeP->m_tableP  = 0;
							pcodeP->m_u4      = sort_atP->from;
							hash_pcode(pcodeP);
							++pcodeP;
						}

						assert(((u1T *) pcodeP) >= g_memory_startP);
						assert(((u1T *) (pcodeP + 1)) <=  g_memory_startP + g_memory);
						pcodeP->m_lineno  = current_lineno;
						pcodeP->m_offset  = (pcodeP - (pcodeT *) memoryP);
						pcodeP->m_pc      = pc;
						pcodeP->m_pcode   = (u1T) opcodeP->pcode;
					
						if (flags & IX) {
							pcodeP->m_u4 = u4;
						} else {
							pcodeP->m_s4 = s4;
					}	}
					
					// Pushing Constants
					// http://sunsite.ee/java/vmspec/vmspec-14.html#HEADING14-0

					switch (pcode) {
					case ILOAD:
					case ILOAD_0:
					case ILOAD_1:
					case ILOAD_2:
					case ILOAD_3:
					case LLOAD:
					case LLOAD_0:
					case LLOAD_1:
					case LLOAD_2:
					case LLOAD_3:
					case FLOAD:
					case FLOAD_0:
					case FLOAD_1:
					case FLOAD_2:
					case FLOAD_3:
					case DLOAD:
					case DLOAD_0:
					case DLOAD_1:
					case DLOAD_2:
					case DLOAD_3:
					case ALOAD:
					case ALOAD_0:
					case ALOAD_1:
					case ALOAD_2:
					case ALOAD_3:

					case ISTORE:
					case ISTORE_0:
					case ISTORE_1:
					case ISTORE_2:
					case ISTORE_3:
					case LSTORE:
					case LSTORE_0:
					case LSTORE_1:
					case LSTORE_2:
					case LSTORE_3:
					case FSTORE:
					case FSTORE_0:
					case FSTORE_1:
					case FSTORE_2:
					case FSTORE_3:
					case DSTORE:
					case DSTORE_0:
					case DSTORE_1:
					case DSTORE_2:
					case DSTORE_3:
					case ASTORE:
					case ASTORE_0:
					case ASTORE_1:
					case ASTORE_2:
					case ASTORE_3:
					{
						varT	*varP;
						
						varP   = (varT *) Align(extraP);
						extraP = (u1T *) (varP + 1);
						if (load) {
							varP->m_number       = u4;
							varP->m_nameP        = 0;
							varP->m_descriptorP  = 0;

							for (localvariableP = localvariablesP; localvariableP < localvariable_endP; ++localvariableP) {
								if (localvariableP->m_index == u4 && localvariableP->m_start_pc <= pc+1 && localvariableP->m_end_pc >= pc) {
									varP->m_nameP       = localvariableP->m_nameP;
									varP->m_descriptorP = localvariableP->m_descriptorP;
									break;
							}	}
							pcodeP->m_tableP     = (u4T *) varP;
						}
						break;
					}
					case IINC:		// iinc
					{
						iincT	*varP;
						
						varP   = (iincT *) Align(extraP);
						extraP = (u1T *) (varP + 1);
						
						if (wide_index) {
							image_at2P = getU2(image_at2P, &u2);
							s2 = ((signed short) u2);
							wide_index = 0;
						} else {
							s2 = *((signed char *) image_at2P);
							++image_at2P;
						}

						if (load) {
							varP->m_number       = u4;
							varP->m_nameP        = 0;
							varP->m_descriptorP  = 0;
							varP->m_s4           = s2;

							for (localvariableP = localvariablesP; localvariableP < localvariable_endP; ++localvariableP) {
								if (localvariableP->m_index == u4 && localvariableP->m_start_pc <= pc+1 && localvariableP->m_end_pc >= pc) {
									varP->m_nameP       = localvariableP->m_nameP;
									varP->m_descriptorP = localvariableP->m_descriptorP;
									break;
							}	}
							pcodeP->m_tableP     = (u4T *) varP;
						}
						break;
					}
					case GETSTATIC:	
					case GETFIELD:
					case PUTSTATIC:
					case PUTFIELD:
					{
						fieldrefT	*fieldrefP;
						
						fieldrefP = (fieldrefT *) Align(extraP);
						extraP    = (u1T *) (fieldrefP + 1);

						if (load) {
							cp_infoT	*cpP;

							assert(u4 > 0 && u4 < m_constant_pool_count);
							cpP = m_constant_poolP + u4;
							assert(cpP->tag == CONSTANT_Fieldref);
							fieldrefP->m_classnameP   = get_classname(cpP->u.ref.index1);
							locate(fieldrefP->m_classnameP);
							u4 = cpP->u.ref.index2;
							assert(u4 > 0 && u4 < m_constant_pool_count);
							cpP = m_constant_poolP + u4;
							assert(cpP->tag == CONSTANT_NameAndType);
							fieldrefP->m_nameP       = get_Utf8(cpP->u.ref.index1);
							fieldrefP->m_descriptorP = get_Utf8(cpP->u.ref.index2);
							pcodeP->m_tableP = (u4T *) fieldrefP;
						}
						break;
					}
					case INVOKEVIRTUAL:
					case INVOKESPECIAL:
					case INVOKESTATIC:
					{
						methodrefT	*methodrefP;
						
						methodrefP = (methodrefT *) Align(extraP);
						extraP    = (u1T *) (methodrefP + 1);

						if (load) {
							cp_infoT	*cpP;

							assert(u4 > 0 && u4 < m_constant_pool_count);
							cpP = m_constant_poolP + u4;
							assert(cpP->tag == CONSTANT_Methodref);
							methodrefP->m_classnameP   = get_classname(cpP->u.ref.index1);
							locate(methodrefP->m_classnameP);
							u4 = cpP->u.ref.index2;
							assert(u4 > 0 && u4 < m_constant_pool_count);
							cpP = m_constant_poolP + u4;
							assert(cpP->tag == CONSTANT_NameAndType);
							methodrefP->m_nameP       = get_Utf8(cpP->u.ref.index1);
							methodrefP->m_descriptorP = get_Utf8(cpP->u.ref.index2);
							pcodeP->m_tableP          = (u4T *) methodrefP;
						}
						break;
					}
					case INVOKEINTERFACE:
					{
						interfacerefT	*methodrefP;
						
						methodrefP = (interfacerefT *) Align(extraP);
						extraP    = (u1T *) (methodrefP + 1);

						if (load) {
							cp_infoT	*cpP;

							assert(u4 > 0 && u4 < m_constant_pool_count);
							cpP = m_constant_poolP + u4;
							assert(cpP->tag == CONSTANT_InterfaceMethodref);
							methodrefP->m_classnameP   = get_classname(cpP->u.ref.index1);
							locate(methodrefP->m_classnameP);
							u4 = cpP->u.ref.index2;
							assert(u4 > 0 && u4 < m_constant_pool_count);
							cpP = m_constant_poolP + u4;
							assert(cpP->tag == CONSTANT_NameAndType);
							methodrefP->m_nameP       = get_Utf8(cpP->u.ref.index1);
							methodrefP->m_descriptorP = get_Utf8(cpP->u.ref.index2);
							methodrefP->m_args        = *image_at2P;
							pcodeP->m_tableP          = (u4T *) methodrefP;
						}
						++image_at2P;
						assert(!*image_at2P); // 2nd byte reserved (should always be 0)
						++image_at2P;
						break;
					}

					// Managing arrays
					// http://sunsite.ee/java/vmspec/vmspec-18.html#HEADING18-0

					case MULTIANEWARRAY:
					{
						arrayrefT	*arrayrefP = (arrayrefT *) Align(extraP);

						extraP = (u1T *) (arrayrefP + 1);
						if (load) {
							arrayrefP->m_classnameP = get_classname(u4);
							locate(arrayrefP->m_classnameP);
							arrayrefP->m_dimensions = *image_at2P;
							pcodeP->m_tableP        = (u4T *) arrayrefP;
						}
						++image_at2P;
						break;
					}
					case ANEWARRAY:
					case NEW:
					case CHECKCAST:
					case INSTANCEOF:
					{
						classrefT	*classrefP = (classrefT *) Align(extraP);

						extraP = (u1T *) (classrefP + 1);
						if (load) {
							classrefP->m_classnameP = get_classname(u4);
							locate(classrefP->m_classnameP);
							pcodeP->m_tableP = (u4T *) classrefP;
						}
						break;
					}
					case LDC:
					case LDC_W:
					case LDC2_W:
					{
						if (load) {
							assert(u4 > 0 && u4 < m_constant_pool_count);
							pcodeP->m_tableP = (u4T *) (m_constant_poolP + u4);
						}
						break;
					}
					case IFEQ:
					case IFNE:
					case IFLT:
					case IFGE:
					case IFGT:
					case IFLE:
					case IF_ICMPEQ:
					case IF_ICMPNE:
					case IF_ICMPLT:
					case IF_ICMPGE:
					case IF_ICMPGT:
					case IF_ICMPLE:
					case IF_ACMPEQ:
					case IF_ACMPNE:
					case IFNULL:
					case IFNOTNULL:
					case GOTO:
					case GOTO_W:
					case JSR:
					case JSR_W:
					{
						if (!load) {
							Csort::append_sort(method_number, pc + s4, pc, LABEL);
						}
						break;
					} 
					case TABLESWITCH:	// tableswitch
					{
						u4T	def;
						s4T	low, high;
						s4T	i;
						u4T	*tableP;

						while ((image_at2P-startCodeP) & 3) {
							assert(!*image_at2P);
							++image_at2P;
						}
						extraP     = (u1T *) Align(extraP);
						image_at2P = getU4(image_at2P,         &def);	// default
						image_at2P = getU4(image_at2P, (u4T *) &low);	// low
						image_at2P = getU4(image_at2P, (u4T *) &high);	// high
						assert(high >= low);
						if (load) {
							tableP = (u4T *) extraP;
							pcodeP->m_tableP = tableP;
							AssertCode(def);
							*tableP++ = def;
							*tableP++ = (u4T) low;
							*tableP++ = (u4T) high;
							for (i = low; i <= high; ++i) {
								image_at2P = getU4(image_at2P, tableP);	// jump table
								AssertCode(*tableP);
								++tableP;
							}
							extraP = (u1T *) tableP;
						} else {
							Csort::append_sort(method_number, pc + (int) def, pc, LABEL);
							for (i = low; i <= high; ++i) {
								image_at2P = getU4(image_at2P, &u4); // jump table
								Csort::append_sort(method_number, pc + (int) u4, pc, LABEL);
							}
							extraP     += (high - low + 4) * sizeof(u4T);	// Jump table plus three initial values
						}
						break;
					}
					case LOOKUPSWITCH:
					{
						u4T	def, pairs, i;
						u4T	*tableP;

						while ((image_at2P - startCodeP) & 3) {
							assert(!*image_at2P);
							++image_at2P;
						}
						extraP     = (u1T *) Align(extraP);
						image_at2P = getU4(image_at2P, &def);	// default
						image_at2P = getU4(image_at2P, &pairs);	// number of pairs
						assert(pairs > 0);
						if (load) {
							pcodeP->m_tableP = tableP = (u4T *) extraP;
							*tableP++ = def;
							*tableP++ = pairs;
							for (i = 0; i < pairs; ++i) {
								image_at2P = getU4(image_at2P, tableP);	// match
								++tableP;
								image_at2P = getU4(image_at2P, tableP);	// jump
								AssertCode(*tableP);
								++tableP;
							}
							extraP      = (u1T *) tableP;
						} else {
							Csort::append_sort(method_number, pc + (int) def, pc, LABEL);
							for (i = 0; i < pairs; ++i) {
								image_at2P += sizeof(u4T);				// match
								image_at2P = getU4(image_at2P, &u4);	// jump
								Csort::append_sort(method_number, pc + (int) u4, pc, LABEL);
							}	
							extraP     += ((pairs + 1) << 1) * sizeof(u4T);
						}
						break;
					}
					case WIDE:					// wide
						wide_index = 1;
						--pcodeP;				// Don't include in the assembler
						continue;
					}
					if (load) {
						hash_pcode(pcodeP);
				}	}
				assert(image_at2P == endCodeP);
				memoryP = (u1T *) pcodeP;
				if (load) {
					methodP->m_pcode_count = pcodeP - methodP->m_pcodeP;
				}
			}
		}
	}

	// Classfile attributes follow

	image_atP = getU2(image_atP, &attributes_count);
	for (j = 0; j < attributes_count; ++j) {
		image_atP  = getU2(image_atP, &attribute_name_index);
		AssertIndex(attribute_name_index);
		image_atP  = getU4(image_atP, &attribute_length);
		nameP      = get_Utf8(attribute_name_index);
		image_at1P = image_atP;
		image_atP += attribute_length;

		if (load) {
			if (!unicodecmp(nameP, synthetic)) {
				assert(attribute_length == 0);
				m_flags |= SyntheticX;
				continue;
			}
			if (!unicodecmp(nameP, deprecated)) {
				assert(attribute_length == 0);
				m_flags |= DeprecatedX;
				continue;
			}
			if (!unicodecmp(nameP, sourceFile)) {
				assert(attribute_length == 2);
				if (m_source_name) {
					fprintf(stderr, "Multiple source declarations in %s\n", m_filenameP);
					error_exit(1);
				}
				getU2(image_at1P, &m_source_name);
				AssertIndex(m_source_name);
				continue;
			}
			if (!unicodecmp(nameP, enclosingMethod)) {
				assert(attribute_length == 4);
				image_at1P = getU2(image_at1P, &m_enclosingMethod.m_class_index);
				AssertIndex(m_enclosingMethod.m_class_index);
				image_at1P = getU2(image_at1P, &m_enclosingMethod.m_method_index);
				AssertIndex(m_enclosingMethod.m_method_index);
				continue;
			}
		}
		if (!unicodecmp(nameP, innerClasses)) {
			innerclassT	*innerclassP;
			u2T			innerclasses_count;
			
			image_at1P           = getU2(image_at1P, &innerclasses_count);
			m_innerclasses_count = innerclasses_count;
			m_innerclassesP      = innerclassP = (innerclassT *) Align(memoryP);
			if (load) {
				for (i = 0; i < innerclasses_count; ++i) {
					image_at1P = getU2(image_at1P, &innerclassP->m_inner_class_info_index);
					AssertIndex(innerclassP->m_inner_class_info_index);
					image_at1P = getU2(image_at1P, &innerclassP->m_outer_class_info_index);
					AssertIndex(innerclassP->m_outer_class_info_index);
					image_at1P = getU2(image_at1P, &innerclassP->m_inner_class_name_index);
					AssertIndex(innerclassP->m_inner_class_name_index);
					image_at1P = getU2(image_at1P, &innerclassP->m_inner_class_access_flags);
					++innerclassP;
				}
				assert (image_at1P == image_atP);
			} else {
				innerclassP += innerclasses_count;
			}
			memoryP = (u1T *) innerclassP;
			continue;
		}
	}
	
	assert((image_atP - g_imageP) == g_image_size);
	if (!load) {
		g_memory = (memoryP - (u1T *) 0) + Csort::g_labels_count * sizeof(pcodeT);
		g_extra  = extraP  - (u1T *) 0;
		g_memory_startP = (u1T *) xmalloc(g_memory);
		g_extra_startP  = (u1T *) xmalloc(g_extra);
		Csort::sort();
	} else {
		assert((memoryP - g_memory_startP) == g_memory);
		assert((extraP  - g_extra_startP)  == g_extra);
	}
	return(0);
fail:
	return(-1);
}

#ifdef WIN32

#include <process.h>

static int
compile(char *pathP)
{
	char		*sourceP, *P, *P1;
	intptr_t	ret;
	int			i;

	sourceP = pathP;
	P       = strrchr(sourceP, '/');
	P1      = strchr(sourceP, '\\');
	if (P1 > P) {
		P = P1;
	}
	if (!P) {
		P = sourceP;
	}
	P       = strchr(P, '$');
	if (P) {
		int	lth = P-pathP;

		sourceP = (char *) alloca(lth+6);
		strncpy(sourceP, pathP, lth);
		strcpy(sourceP+lth, ".java");
	}	
	g_java[4] = sourceP;

	if (g_verbose) {
		for (i = 0; P = g_java[i]; ++i) {
			fputs(P, stderr);
			fputc(' ', stderr);
		}
		fputc('\n', stderr);
	}

	ret = _spawnvpe(_P_WAIT, g_java[0], g_java+1, g_envPP);
	if (ret) {
		if (!g_verbose) {
			for (i = 0; P = g_java[i]; ++i) {
				fputs(P, stderr);
				fputc(' ', stderr);
		}	}
		fprintf(stderr, "returned %d\n", ret);
	} 
	return(ret);
}
#else
#include <sys/wait.h>

static int
compile(char *pathP)
{
	char		*sourceP, *P;
	int			ret, ret1, pid, status;
	int			i;

	sourceP = pathP;
	P       = strrchr(sourceP, '/');
	if (!P) {
		P = sourceP;
	}
	P       = strchr(P, '$');
	if (P) {
		int	lth = P-pathP;

		sourceP = (char *) alloca(lth+6);
		strncpy(sourceP, pathP, lth);
		strcpy(sourceP+lth, ".java");
	}	
	g_java[4] = sourceP;

	if (g_verbose) {
		for (i = 0; P = g_java[i]; ++i) {
			fputs(P, stderr);
			fputc(' ', stderr);
		}
		fputc('\n', stderr);
	}

	ret  = fork();
	ret1 = -2;

	switch (ret) {
	case 0:
		execvp(g_java[0], g_java);
		fprintf(stderr,"Return from execve %s\n", g_java[0]);
		error_exit(1);
	case -1:
		fprintf(stderr,"Fork failed\n");
		break;
	default:
		pid = wait(&status);
		if (pid != ret) {
			fprintf(stderr,"Wait for %s to finish failed\n", g_java[0]);
		} else if (!WIFEXITED(status)) {
			fprintf(stderr,"%s did not terminate normally\n", g_java[0]);
		} else {
			ret1 = WEXITSTATUS(status);
	}	}
	if (ret1) {
		if (!g_verbose) {
			for (i = 0; P = g_java[i]; ++i) {
				fputs(P, stderr);
				fputc(' ', stderr);
		}	}
		fprintf(stderr, "returned %d\n", ret1);
	} 
	return(ret1);
}
#endif

int
readclassfile(char *pathP, Cclass *classP)
{
	struct	stat 	stats;
	char			*bufferP;
	int				lth;
	Cclass			*class1P;
	int				load, ret;

	if (stat(pathP, &stats)) {
		fprintf(stdoutF, "Can't stat %s%s<br>\n", (classP ? "dynamic " : ""), pathP);
		if (classP) {
			return(1);
		}
		return(-1);
	}

	lth = strlen(pathP);
	if (lth > 5 && !strcmp(pathP+lth-5,".java")) {
		if (!g_java[0]) {
			fprintf(stderr,"Use -j option if you want to compile %s\n", pathP);
			return(-1);
		}
		fputs("Compiled ", stdoutF);
		fputs(pathP, stdoutF);
		ret = compile(pathP);
		if (ret) {
			fprintf(stdoutF, " failed ret=%d", ret);
		}
		fputc('\n', stdoutF);
		
		bufferP = (char *) alloca(lth+2);
		strcpy(bufferP, pathP);
		strcpy(bufferP+lth-4,"class");
		pathP = bufferP;

		if (stat(pathP, &stats)) {
			if (classP) {
				return(1);
			}
			fprintf(stderr, "Can't stat %s\n", pathP);
			return(-1);
		}
	}

	ret = read_image(pathP, &stats);
	if (ret != 0) {
		goto done;
	}

	if (classP) {
		fputs("Dynamic ", stdoutF);
	}
	fputs(pathP, stdoutF);
	fputs("\n<br>\n", stdoutF);

	class1P = classP;
	if (!class1P) {
		class1P = new Cclass();
	}

	for (load = 0; load < 2; ++load) {
		ret = class1P->load_constant_pool(pathP, load);
		if (ret != 0) {
			goto done;
	}	}

	for (load = 0; load < 2; ++load) {
		ret = class1P->load_tables(load);
		if (ret != 0) {
			goto done;
	}	}

	if (g_verbose) {
		if (classP) {
			fputs("Dynamically processed", stderr);
		} else {
			fputs("Processed", stderr);
		}
		fprintf(stderr, " \"%s\"\n", pathP);
	}
done:
	return(ret);
}

void
pseudo_source(void)
{
	int			i, ret;
	Cclass		*classP;
	char		*classpathP;
	int			repeat;
	char		path[2048];
	char		*pathP, *P;
	u2T			*P2;
	char		c;

	classpathP = g_classpathP;
	if (!classpathP) {
		classpathP = ".";
	}
	do {
		repeat = 0;
		for (i = 0; i < HASHSIZE; ++i) {
			for (classP = Cclass::g_hash[i]; classP; classP = classP->m_hashP) {
				if (!(classP->m_flags & ProcessedX)) {
					if (g_dynamic) {
						pathP = path;
						for (P = classpathP;; ++P) {
							switch (c = *P) {
							case 0:
							case ';':
							case ':':
							case ',':
								*pathP++ = '/';
								break;
							default:
								*pathP++ = c;
								continue;
							}
							for (P2 = classP->m_nameP; *pathP++ = (u1T) *P2++; );

							--pathP;
							ret = 1;
							if (g_java[0]) {
								strcpy(pathP, ".java");
								ret = readclassfile(path, classP);
							} 
							if (ret == 1) {
								strcpy(pathP, ".class");
								ret = readclassfile(path, classP);
							}
							switch (ret) {
							case 0:
								repeat = 1;
								break;
							case 1:
								if (c && P[1]) {
									// Wrong item on class path used; move to next item on class path
									pathP = path;
									continue;
								}
								break;
							}
							break;
					}	}
					classP->m_flags |= ProcessedX;
		}	}	}
	} while (repeat);
}


